@@ -29,6 +29,8 @@ import fnmatch
 import io
 import smtplib
 
+from pathlib import Path
+
 from email.message import EmailMessage
 from fcntl import lockf, LOCK_EX, LOCK_UN, LOCK_NB
 
@@ -52,6 +54,15 @@ def log_errors(fullpath, cmdargs, lines):
                 break
 
 
+def gen_preload_bundle(fullpath, config):
+    outdir = config['fsck'].get('preload_bundle_outdir')
+    Path(outdir).mkdir(parents=True, exist_ok=True)
+    bname = '%s.bundle' % os.path.basename(fullpath)[:-4]
+    args = ['bundle', 'create', os.path.join(outdir, bname), '--all']
+    logger.info(' bundling: %s', bname)
+    grokmirror.run_git_command(fullpath, args)
+
+
 def get_blob_set(fullpath):
     bset = set()
     size = 0
@@ -715,6 +726,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
     queued = 0
     logger.info('Analyzing %s (%s repos)', toplevel, len(status))
     stattime = time.time()
+    baselines = [x.strip() for x in config['fsck'].get('baselines', '').split('\n')]
     for fullpath in list(status):
         # Give me a status every 5 seconds
         if time.time() - stattime >= 5:
@@ -758,7 +770,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
 
         if not altdir and not os.path.exists(os.path.join(fullpath, 'grokmirror.do-not-objstore')):
             # Do we match any obstdir repos?
-            obstrepo = grokmirror.find_best_obstrepo(fullpath, obst_roots)
+            obstrepo = grokmirror.find_best_obstrepo(fullpath, obst_roots, toplevel, baselines)
             if obstrepo:
                 obst_changes = True
                 # Yes, set ourselves up to be using that obstdir
@@ -825,7 +837,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
             # We have an alternates repo, but it's not an objstore repo
             # Probably left over from grokmirror-1.x
             # Do we have any matching obstrepos?
-            obstrepo = grokmirror.find_best_obstrepo(fullpath, obst_roots)
+            obstrepo = grokmirror.find_best_obstrepo(fullpath, obst_roots, toplevel, baselines)
             if obstrepo:
                 logger.info('%s: migrating to %s', gitdir, os.path.basename(obstrepo))
                 if altdir not in fetched_obstrepos:
@@ -988,7 +1000,6 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
     analyzed = 0
     queued = 0
     logger.info('Analyzing %s (%s repos)', obstdir, len(obstrepos))
-    baselines = [x.strip() for x in config['fsck'].get('baselines', '').split('\n')]
     islandcores = [x.strip() for x in config['fsck'].get('islandcores', '').split('\n')]
     stattime = time.time()
     for obstrepo in obstrepos:
@@ -1040,9 +1051,11 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
         my_remotes = grokmirror.list_repo_remotes(obstrepo, withurl=True)
         # Use the first child repo as our "reference" entry in manifest
         refrepo = None
-        set_baseline = False
+        # Use for the alternateRefsPrefixes value
+        baseline_refs = set()
         set_islandcore = False
         new_islandcore = False
+        valid_virtrefs = set()
         for virtref, childpath in my_remotes:
             # Is it still relevant?
             if childpath not in amap[obstrepo]:
@@ -1050,6 +1063,7 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                 grokmirror.remove_from_objstore(obstrepo, childpath)
                 logger.info('%s: removed remote %s (no longer used)', os.path.basename(obstrepo), childpath)
                 continue
+            valid_virtrefs.add(virtref)
 
             # Does it need fetching?
             fetch = True
@@ -1074,24 +1088,11 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                 continue
 
             # Do we need to set any alternateRefsPrefixes?
-            if not set_baseline:
-                is_baseline = False
-                for baseline in baselines:
-                    # Does this repo match a baseline
-                    if fnmatch.fnmatch(gitdir, baseline):
-                        is_baseline = True
-                        break
-                if is_baseline:
-                    set_baseline = True
-                    refpref = 'refs/virtual/%s/heads/' % virtref
-                    # Go through all remotes and set their alternateRefsPrefixes
-                    for s_virtref, s_childpath in my_remotes:
-                        # is it already set to that?
-                        entries = grokmirror.get_config_from_git(s_childpath, r'core\.alternate*')
-                        if entries.get('alternaterefsprefixes') != refpref:
-                            s_gitdir = '/' + os.path.relpath(s_childpath, toplevel)
-                            logger.info(' reconfig: %s (baseline to %s)', s_gitdir, virtref)
-                            grokmirror.set_git_config(s_childpath, 'core.alternateRefsPrefixes', refpref)
+            for baseline in baselines:
+                # Does this repo match a baseline
+                if fnmatch.fnmatch(gitdir, baseline):
+                    baseline_refs.add('refs/virtual/%s/heads/' % virtref)
+                    break
 
             # Do we need to set islandCore?
             if not set_islandcore:
@@ -1119,11 +1120,41 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
 
             manifest[gitdir]['forkgroup'] = os.path.basename(obstrepo[:-4])
 
+        if len(baseline_refs):
+            # sort the list, so we have deterministic value
+            br = list(baseline_refs)
+            br.sort()
+            refpref = ' '.join(br)
+            # Go through all remotes and set their alternateRefsPrefixes
+            for s_virtref, s_childpath in my_remotes:
+                # is it already set to that?
+                entries = grokmirror.get_config_from_git(s_childpath, r'core\.alternate*')
+                if entries.get('alternaterefsprefixes') != refpref:
+                    s_gitdir = '/' + os.path.relpath(s_childpath, toplevel)
+                    logger.info(' reconfig: %s (baseline)', s_gitdir)
+                    grokmirror.set_git_config(s_childpath, 'core.alternateRefsPrefixes', refpref)
         repack_requested = False
         if os.path.exists(os.path.join(obstrepo, 'grokmirror.repack')):
             repack_requested = True
 
-        if obstrepo not in status or new_islandcore or repack_requested:
+        # Go through all our refs and find all stale virtrefs
+        args = ['for-each-ref', '--format=%(refname)', 'refs/virtual/']
+        trimmed_virtrefs = set()
+        ecode, out, err = grokmirror.run_git_command(obstrepo, args)
+        if ecode == 0 and out:
+            for line in out.split('\n'):
+                chunks = line.split('/')
+                if len(chunks) < 3:
+                    # Where did this come from?
+                    logger.debug('Weird ref %s in objstore repo %s', line, obstrepo)
+                    continue
+                virtref = chunks[2]
+                if virtref not in valid_virtrefs and virtref not in trimmed_virtrefs:
+                    logger.info('     trim: stale virtref %s', virtref)
+                    grokmirror.objstore_trim_virtref(obstrepo, virtref)
+                    trimmed_virtrefs.add(virtref)
+
+        if obstrepo not in status or new_islandcore or trimmed_virtrefs or repack_requested:
             # We don't use obstrepo fingerprints, so we set it to None
             status[obstrepo] = {
                 'lastcheck': 'never',
@@ -1200,6 +1231,9 @@ def fsck_mirror(config, force=False, repack_only=False, conn_only=False,
                     status[fullpath]['lastfullrepack'] = todayiso
                     status[fullpath]['lastcheck'] = todayiso
                     status[fullpath]['nextcheck'] = nextcheck.strftime('%F')
+                    # Do we need to generate a preload bundle?
+                    if config['fsck'].get('preload_bundle_outdir') and grokmirror.is_obstrepo(fullpath, obstdir):
+                        gen_preload_bundle(fullpath, config)
                     logger.info('     next: %s', status[fullpath]['nextcheck'])
             else:
                 logger.warning('Repacking %s was unsuccessful', fullpath)
